"""
    Common data types for LabVIEW resources
"""


import struct
import enum
from functools import total_ordering
import re


ESCAPED_PATTERN = re.compile(b"\\\\x[0-9a-f][0-9a-f]")
ESCAPE_CHARS = re.compile(b"[^ -[\\]-~]")


def unescape_bytes(matchobj) -> bytes:
    """unescape non-ascii bytes"""
    return UInt8(int(matchobj.group(0)[2:], 16)).to_bytes()


def escape_bytes(matchobj) -> bytes:
    """escape non-ascii bytes"""
    return b"\\x" + matchobj.group(0).hex().encode("ascii")


class Description:
    """Base class for all descriptions"""

    def to_string(self) -> str:
        """Get a string representation"""
        return ""

    def size(self):
        """no size for an empty description"""
        return 0

    # pylint: disable=unused-argument
    def from_bytes(self, data: bytes, offset: int = 0):
        """no bytes to read"""
        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        return b""

    def __str__(self):
        return self.to_string()


class Endian(enum.Enum):
    """integer endian type"""

    BIG = ">"
    LITTLE = "<"


class IntSize(enum.Enum):
    """struct pack characters"""

    INT8 = "b"
    INT16 = "h"
    INT32 = "i"


@total_ordering
class Integer(Description):
    """Integer type"""

    SIZES = {
        IntSize.INT8: 1,
        IntSize.INT16: 2,
        IntSize.INT32: 4,
    }

    def __init__(
        self,
        value: int = None,
        endian: Endian = Endian.BIG,
        byte_count: IntSize = IntSize.INT32,
        signed: bool = False,
    ):
        assert value is None or signed or value >= 0
        self.value = value
        self.endian = endian
        self.signed = signed
        self.byte_count = byte_count

    def size(self) -> int:
        """Get the size of the bytes representation"""
        return Integer.SIZES[self.byte_count]

    def __struct_description(self):
        return self.endian.value + (
            self.byte_count.value if self.signed else self.byte_count.value.upper()
        )

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        assert len(data) >= offset + self.size(), [self.size(), offset, data]
        self.value = struct.unpack(
            self.__struct_description(), data[offset : offset + self.size()]
        )[0]
        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        assert self.value is None or self.signed or self.value >= 0
        return struct.pack(self.__struct_description(), self.value)

    def from_value(self, description: any):
        """Get a Python basic type to represent this value"""
        self.value = description
        assert self.value is None or self.signed or self.value >= 0
        return self

    def to_value(self) -> any:
        """restore from a basic python type"""
        assert self.value is None or self.signed or self.value >= 0
        return self.value

    def to_string(self) -> str:
        """Get a string representation"""
        return str(self.value)

    def __repr__(self) -> str:
        return (
            f"Integer({self.to_string()}, {self.endian}, "
            + f"signed={self.signed}, {self.byte_count})"
        )

    def __lt__(self, other) -> bool:
        if isinstance(other, int):
            return self.value < other

        if isinstance(other, bytes):
            return self.value < struct.unpack(self.__struct_description(), other)[0]

        return self.value < other.value

    def __eq__(self, other) -> bool:
        if isinstance(other, int):
            return self.value == other

        if isinstance(other, bytes):
            return self.value == struct.unpack(self.__struct_description(), other)[0]

        return self.value == other.value

    def __add__(self, other) -> int:
        if isinstance(other, int):
            return self.value + other

        return self.value + other.value

    def __hash__(self):
        return self.value


class UInt32(Integer):
    """unsigned 32-bit integer"""

    def __init__(self, value: int = None, endian: Endian = Endian.BIG):
        super().__init__(value, endian, IntSize.INT32, signed=False)


class UInt16(Integer):
    """unsigned 16-bit integer"""

    def __init__(self, value: int = None, endian: Endian = Endian.BIG):
        super().__init__(value, endian, IntSize.INT16, signed=False)


class UInt8(Integer):
    """unsigned 16-bit integer"""

    def __init__(self, value: int = None, endian: Endian = Endian.BIG):
        super().__init__(value, endian, IntSize.INT8, signed=False)


class Bytes(Description):
    """Collection of Bytes"""

    def __init__(self, value: bytes = None, byte_count: int = 0):
        if isinstance(value, str):
            value = value.encode("ascii")

        elif value is None and byte_count > 0:
            value = b"\x00" * byte_count

        self.value = value

    def size(self) -> int:
        """Get the size of the bytes representation"""
        return len(self.value)

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        assert self.value is not None, "You must specify byte_count in constructor"
        self.value = data[offset : offset + self.size()]
        assert len(self.value) == self.size()
        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        return self.value

    def from_value(self, description: any):
        """Get a Python basic type to represent this value"""
        if description is None:
            self.value = description
        else:
            self.value = ESCAPED_PATTERN.sub(
                unescape_bytes, description.encode("ascii")
            )

        return self

    def to_value(self) -> any:
        """restore from a basic python type"""
        if self.value is None:
            return None

        return ESCAPE_CHARS.sub(escape_bytes, self.value).decode("ascii")

    def to_string(self):
        """Get a string representation"""
        return self.to_value()

    def __len__(self):
        return self.size()

    def __repr__(self):
        return f"Bytes('{self.to_string()}')"

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return self.to_string() == other

        if isinstance(other, bytes):
            return self.value == other

        return self.value == other.value


class FourCharCode(Bytes):
    """Four character identifier"""

    def __init__(self, value: bytes = None):
        assert value is None or len(value) == 4
        super().__init__(value, byte_count=4)

    def __repr__(self):
        return f"FourCharCode('{self.to_string()}')"


class PString(Bytes):
    """length-prefixed string"""

    def __init__(
        self,
        value: bytes = None,
        pad_to: IntSize = IntSize.INT8,
        prefix_size: IntSize = IntSize.INT8,
    ):
        assert value is None or len(value) < 2 ** (8 * Integer.SIZES[prefix_size])
        super().__init__(value)
        self.pad_to = Integer.SIZES[pad_to]
        self.prefix_size = prefix_size

    def size(self) -> int:
        """Get the size of the bytes representation"""
        return (
            int((Integer.SIZES[self.prefix_size] + super().size() - 1) / self.pad_to)
            + 1
        ) * self.pad_to

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        assert len(data) >= offset + Integer(byte_count=self.prefix_size).size(), [
            offset,
            Integer(byte_count=self.prefix_size).size(),
            data,
        ]
        byte_count = Integer(byte_count=self.prefix_size, signed=False)
        byte_count.from_bytes(data, offset)
        start = offset + Integer.SIZES[self.prefix_size]
        end = start + byte_count.value
        self.value = data[start:end]
        assert len(self.value) == byte_count.value, [
            len(self.value),
            byte_count,
            data[offset:],
        ]
        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        prefix = Integer(
            len(self.value), byte_count=self.prefix_size, signed=False
        ).to_bytes()
        value = prefix + self.value
        return value + b"\x00" * (self.size() - len(value))

    def __repr__(self):
        return f"PString('{self.to_string()}')"


class PString16(PString):
    """16-bit length prefixed string"""

    def __init__(self, value: bytes = None, pad_to: IntSize = IntSize.INT8):
        super().__init__(value, pad_to, IntSize.INT16)


class PStringPad2(PString):
    """16-bit length prefixed string"""

    def __init__(self, value: bytes = None, prefix_size: IntSize = IntSize.INT8):
        super().__init__(value, pad_to=IntSize.INT16, prefix_size=prefix_size)


class Structure(Description):
    """A collection of heterogeneous, named data"""

    def __init__(self, *value):
        self.__fields = list(value[0::2])
        assert all(isinstance(n, str) for n in self.__fields)
        assert all(isinstance(v, Description) for v in value[1::2])
        self.__dict__.update(dict(zip(self.__fields, value[1::2])))

    def size(self):
        """Get the size of the bytes representation"""
        return sum(self.__dict__[n].size() for n in self.__fields)

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        for name in self.__fields:
            self.__dict__[name].from_bytes(data, offset)
            offset += self.__dict__[name].size()

        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        return b"".join(self.__dict__[n].to_bytes() for n in self.__fields)

    def from_value(self, description: any):
        """Get a Python basic type to represent this value"""
        assert set(description.keys()) <= set(self.__fields)

        for name in self.__fields:
            if name in description:
                self.__dict__[name].from_value(description[name])

        return self

    def to_value(self) -> any:
        """restore from a basic python type"""
        return {n: self.__dict__[n].to_value() for n in self.__fields}

    def to_string(self):
        """Get a string representation"""
        return str(self.to_value())

    def __repr__(self):
        return f"Structure({', '.join(n+'='+repr(self.__dict__[n]) for n in self.__fields)})"

    def __getitem__(self, key):
        for name in self.__fields:
            if name == key:
                return self.__dict__[name]

        return None

    # pylint: disable=protected-access
    def __eq__(self, other):
        if not isinstance(other, Structure) or self.__fields != other.__fields:
            return False

        return all(self.__dict__[n] == other.__dict__[n] for n in self.__fields)


class Array(Description):
    """A list of data of the same type"""

    def __init__(self, data_type, *value, data_count=0):
        assert all(isinstance(v, data_type) for v in value), [data_type, repr(value)]
        assert len(value) == 0 or data_count == 0
        self.data_type = data_type
        self.value = value if value else [data_type() for _ in range(0, data_count)]

    def size(self) -> int:
        """Get the size of the bytes representation"""
        assert self.length() >= 0
        return sum(v.size() for v in self.value)

    def length(self) -> int:
        """Get the number of elements in the array"""
        return len(self.value)

    def set_length(self, data_count: int):
        """Clear the array and fill it with data_count empty data elements"""
        self.value = [self.data_type() for _ in range(0, data_count)]
        return self

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        for value in self.value:
            value.from_bytes(data, offset)
            offset += value.size()

        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        return b"".join(v.to_bytes() for v in self.value)

    def from_value(self, description: any):
        """Get a Python basic type to represent this value"""
        self.value = [self.data_type().from_value(v) for v in description]
        return self

    def to_value(self) -> any:
        """restore from a basic python type"""
        return [v.to_value() for v in self.value]

    def to_string(self):
        """Get a string representation"""
        return str(self.to_value())

    def __repr__(self):
        return f"List({', '.join(repr(v) for v in self.value)})"

    def __len__(self):
        return self.length()

    def __getitem__(self, item_index):
        return self.value[item_index]

    def __eq__(self, other):
        if len(self) != len(other):
            return False

        return all(a == b for a, b in zip(self, other))


class Path(Structure):
    """LabVIEW path type"""

    KNOWN_PATH_FORMATS = ["PTH0", "PTH1", "PTH2"]
    ABSOLUTE = 0
    RELATIVE = 1
    NOTAPATH = 2
    UNC = 3
    NOT_A_PATH_STRING = "\\\\NOT A PATH//"
    TYPE_VALUES = [ABSOLUTE, RELATIVE, NOTAPATH, UNC]
    TYPES = ["absolute", "relative", "not a path", "unc"]
    NEW_TYPES = ["abs ", "rel ", "!pth", "unc "]
    STARTS = {
        ABSOLUTE: "/",
        RELATIVE: "",
        UNC: "//",
    }
    NEW_PATH_PATTERN = re.compile(r"^{([0-9])}")

    def __init__(self):
        super().__init__(
            "format",
            FourCharCode(Path.KNOWN_PATH_FORMATS[0]),
            "byte_count",
            UInt32(FourCharCode().size() + UInt32().size()),
        )
        self.type = UInt16(Path.ABSOLUTE)
        self.count = UInt16(0)
        self.elements = Array(PString)
        assert self.format in Path.KNOWN_PATH_FORMATS, self.format
        assert self.type < len(Path.TYPES)

    def __name_to_str(self, name: bytes) -> str:
        if len(name) == 0:
            return ".."

        if self.format != Path.KNOWN_PATH_FORMATS[-1]:
            value = ESCAPE_CHARS.sub(escape_bytes, name).decode("ascii")
        else:
            value = name.decode("utf-8")

        return value.replace("/", "\\x2f")

    def __str_to_name(self, name: str) -> bytes:
        if name == "..":
            return b""

        value = name.replace("\\x2f", "/")

        if self.format != Path.KNOWN_PATH_FORMATS[-1]:
            return ESCAPED_PATTERN.sub(unescape_bytes, value.encode("ascii"))

        return value.encode("utf-8")

    def size(self):
        """Get the size of the bytes representation"""
        return (
            super().size() + self.type.size() + self.count.size() + self.elements.size()
        )

    def is_type(self, path_format: int) -> bool:
        """determine if this path has the given format"""
        if self.type.value in Path.TYPE_VALUES:
            return self.type.value == path_format

        return self.type.to_string() == Path.NEW_TYPES[path_format]

    def get_type(self) -> int:
        """Get the format int"""
        if isinstance(self.type, Integer):
            assert self.type.value < len(Path.TYPES)
            return self.type.value

        assert self.type.to_string() in Path.NEW_TYPES, [
            self.type.to_string(),
            Path.NEW_TYPES,
            self.type,
        ]
        return Path.NEW_TYPES.index(self.type.to_string())

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        super().from_bytes(data, offset)
        assert self.byte_count >= 4 or self.byte_count == 0

        if self.format.to_string() == Path.KNOWN_PATH_FORMATS[0]:
            string_type = PString
            self.type = Integer(0, byte_count=IntSize.INT16)
            self.count = Integer(0, byte_count=IntSize.INT16)
        else:
            assert self.format.to_string() in Path.KNOWN_PATH_FORMATS, [
                self.format,
                data,
                offset,
                data[offset:],
            ]
            string_type = PString16
            self.type = FourCharCode()
            self.count = Description()

        offset += super().size()
        original_offset = offset
        self.type.from_bytes(data, offset)
        assert self.byte_count >= 4 or self.is_type(Path.ABSOLUTE)
        offset += self.type.size()
        self.count.from_bytes(data, offset)
        offset += self.count.size()
        elements = []
        assert self.byte_count + original_offset <= len(data), [
            self.byte_count,
            original_offset,
            data,
            data[offset:],
        ]

        while offset - original_offset < self.byte_count:
            elements.append(string_type().from_bytes(data, offset))
            offset += elements[-1].size()

        assert self.byte_count in {offset - original_offset, 0}, [
            offset,
            original_offset,
            self,
            data,
        ]
        self.elements = Array(string_type, *elements)
        assert not self.is_type(Path.NOTAPATH) or len(self.elements) == 0
        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        return (
            super().to_bytes()
            + self.type.to_bytes()
            + self.count.to_bytes()
            + self.elements.to_bytes()
        )

    # pylint: disable=attribute-defined-outside-init
    def from_value(self, description: any):
        """Get a Python basic type to represent this value"""
        if description is None or description == Path.NOT_A_PATH_STRING:
            self.format = FourCharCode(Path.KNOWN_PATH_FORMATS[0])
            self.byte_count = UInt32(0 if description is None else self.format.size())
            self.type = UInt16(Path.ABSOLUTE if description is None else Path.NOTAPATH)
            self.count = UInt16(0)
            self.elements = Array(PString)
            return self

        new_path_type = Path.NEW_PATH_PATTERN.match(description)

        if new_path_type:
            description = description[new_path_type.end() :]
            type_index = int(new_path_type.group(1))
            name_type = PString16
        else:
            type_index = 0
            name_type = PString

        self.format = FourCharCode(Path.KNOWN_PATH_FORMATS[type_index])
        path_format = [
            f
            for f in [Path.UNC, Path.ABSOLUTE, Path.RELATIVE]
            if description.startswith(Path.STARTS[f])
        ][0]
        self.type = (
            FourCharCode(Path.NEW_TYPES[path_format])
            if new_path_type
            else UInt16(path_format)
        )
        description = description[len(Path.STARTS[path_format]) :]

        if description:
            elements = [
                name_type(self.__str_to_name(n)) for n in description.split("/")
            ]
        else:
            elements = []

        self.elements = Array(name_type, *elements)
        self.count = Description() if new_path_type else UInt16(len(self.elements))
        self.byte_count = UInt32(len(self.to_bytes()) - super().size())
        return self

    def to_value(self) -> any:
        """restore from a basic python type"""
        if self.is_type(Path.NOTAPATH):
            return Path.NOT_A_PATH_STRING

        if self.byte_count == 0:
            return None

        if self.format.to_string() != Path.KNOWN_PATH_FORMATS[0]:
            value = f"{{{Path.KNOWN_PATH_FORMATS.index(self.format.to_string())}}}"
        else:
            value = ""

        value += Path.STARTS[self.get_type()]
        return value + "/".join(self.__name_to_str(n.value) for n in self.elements)

    def to_string(self) -> str:
        return str(self.to_value())

    def __repr__(self) -> str:
        return (
            f"Path({self.format}[{self.byte_count}] "
            + f"{Path.TYPES[self.get_type()]} #{self.count} {self.elements})"
        )


@total_ordering
class Version(UInt32):
    """LabVIEW version number"""

    DEVELOPMENT = 1
    ALPHA = 2
    BETA = 3
    RELEASE = 4
    PHASES = "?dabf"
    PATTERN = re.compile(r"((\d+)(\.(\d+)(\.(\d+))?)?)(([abdfABDF]|{\d})(\d+)?)?")

    @staticmethod
    def __from_string(version: str) -> (int, int, int, int, int):
        final = Version.PHASES[Version.RELEASE]
        correct_format = Version.PATTERN.match(version)
        assert correct_format, f"Incorrect version string: '{version}'"
        major = int(correct_format.group(2))
        minor = int(correct_format.group(4)) if correct_format.group(4) else 0
        patch = int(correct_format.group(6)) if correct_format.group(6) else 0
        phase_str = correct_format.group(8) if correct_format.group(8) else final
        build = int(correct_format.group(9)) if correct_format.group(9) else 0

        if phase_str.startswith("{") and phase_str.endswith("}"):
            phase = int(phase_str[1:-1])
        else:
            phase = Version.PHASES.find(phase_str)
        return major, minor, patch, phase, build

    @staticmethod
    def __compose(major: int, minor: int, patch: int, phase: int, build: int) -> int:
        assert major <= 99, major
        assert minor <= 9, minor
        assert patch <= 9, patch
        assert build <= 1999, build
        return (
            (int(major / 10) << 28)
            + ((major % 10) << 24)
            + (minor << 20)
            + (patch << 16)
            + (phase << 13)
            + (int(build / 1000) << 12)
            + (int(build % 1000 / 100) << 8)
            + (int(build % 100 / 10) << 4)
            + ((build % 10) << 0)
        )

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        major_or_str: any = None,
        minor: int = None,
        patch: int = None,
        phase: int = None,
        build: int = None,
    ):
        version_value = None

        if major_or_str is not None:
            if isinstance(major_or_str, str):
                assert minor is None
                assert patch is None
                assert phase is None
                assert build is None
                major, minor, patch, phase, build = Version.__from_string(major_or_str)
            elif isinstance(major_or_str, bytes):
                assert minor is None
                assert patch is None
                assert phase is None
                assert build is None
                version_value = int.from_bytes(major_or_str, "big")
            elif major_or_str <= 99:
                major = major_or_str
                minor = minor or 0
                patch = patch or 0
                phase = Version.RELEASE if phase is None else phase
                build = build or 0
            else:
                assert minor is None
                assert patch is None
                assert phase is None
                assert build is None
                version_value = major_or_str

            if not version_value:
                version_value = Version.__compose(major, minor, patch, phase, build)

        super().__init__(version_value)

    def size(self) -> int:
        """Get the size of the bytes representation"""
        return Integer.SIZES[self.byte_count]

    def from_value(self, description: any):
        """Get a Python basic type to represent this value"""
        self.value = Version.__compose(*Version.__from_string(description))
        return self

    def to_value(self) -> any:
        """restore from a basic python type"""
        major = self.major()
        minor = self.minor()
        patch = self.patch()
        phase = self.phase()
        build = self.build()
        version_string = str(major)

        if minor > 0 or patch > 0:
            version_string += f".{minor}"

            if patch > 0:
                version_string += f".{patch}"

        if phase != Version.RELEASE or build > 0:
            if Version.DEVELOPMENT <= phase <= Version.RELEASE:
                phase_str = Version.PHASES[phase]
            else:
                phase_str = f"{{{phase}}}"

            version_string += phase_str

            if build > 0:
                version_string += str(build)
        return version_string

    def to_string(self):
        """Get a string representation"""
        return self.to_value()

    def __repr__(self):
        return f"Version('{self.to_string()}')"

    def __lt__(self, other):
        if isinstance(other, Version):
            return self.value < other.value

        return self.value < Version(other).value

    def __eq__(self, other):
        if isinstance(other, Version):
            return self.value == other.value

        return self.value == Version(other).value

    def __hash__(self):
        return self.value

    def major(self):
        """get the major version"""
        return (self.value >> 28 & 0xF) * 10 + (self.value >> 24 & 0xF)

    def minor(self):
        """get the minor version"""
        return int(self.value >> 20 & 0xF)

    def patch(self):
        """get the patch version"""
        return int(self.value >> 16 & 0xF)

    def phase(self):
        """get the version phase"""
        return self.value >> 13 & 0x7

    def build(self):
        """get the build number"""
        return (
            1000 * (self.value >> 12 & 0x01)
            + 100 * (self.value >> 8 & 0xF)
            + 10 * int(self.value >> 4 & 0xF)
            + (self.value >> 0 & 0xF)
        )



"""
    LabVIEW resource data
"""


import hashlib

from pylavi.data_types import Structure, Array, Bytes, UInt32, UInt16
from pylavi.data_types import PString, Version, Path


class TypePATH(Path):
    """path"""


class TypeDLLP(Path):
    """executable dll path"""


class TypeHLPP(Path):
    """help path"""


class TypeRTMP(Path):
    """VI runtime menu path"""


class TypeLPTH(Array):
    """32-bit-integer-length-prefixed list of paths"""

    def __init__(self):
        super().__init__(Path)

    def size(self) -> int:
        """Get the size of the bytes representation"""
        assert self.length() >= 0
        return UInt32().size() + super().size()

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        data_count = UInt32().from_bytes(data, offset)
        offset += data_count.size()
        self.set_length(data_count.value)
        super().from_bytes(data, offset)
        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        return UInt32(self.length()).to_bytes() + super().to_bytes()

    def __repr__(self):
        return f"TypeLPTH({', '.join(repr(v) for v in self.value)})"


class TypeBDPW(Array):
    """handles 'BDPW' resources types"""

    EMPTY_PASSWORD = "d41d8cd98f00b204e9800998ecf8427e"
    PASSWORD_MD5_INDEX = 0
    MD5_SIZE = 16

    def __init__(self):
        super().__init__(Bytes)

    def __password_hash(self):
        return self.value[TypeBDPW.PASSWORD_MD5_INDEX].value

    def has_password(self):
        """Determines if a password was set"""
        return self.__password_hash().hex().lower() != TypeBDPW.EMPTY_PASSWORD

    def password_matches(self, password: str) -> bool:
        """checks the password to see if it matches"""
        return hashlib.md5(password.encode("ascii")).digest() == self.__password_hash()

    def from_bytes(self, data: bytes, offset: int = 0):
        """Take raw bytes from the file and interpret them.
        offset - the offset in data to start parsing the bytes
        """
        assert (len(data) - offset) % TypeBDPW.MD5_SIZE == 0
        md5_count = int((len(data) - offset) / TypeBDPW.MD5_SIZE)
        self.value = [Bytes(byte_count=TypeBDPW.MD5_SIZE) for _ in range(0, md5_count)]
        return super().from_bytes(data, offset)

    def from_value(self, description: any):
        self.value = [Bytes(bytes.fromhex(v)) for v in description]
        assert all(len(b.value) == TypeBDPW.MD5_SIZE for b in self.value)
        return self

    def to_value(self) -> any:
        return [v.value.hex().lower() for v in self.value]

    def __repr__(self) -> str:
        return f"TypeBDPW({self.to_string()})"


# pylint: disable=no-member
class TypeLVSR(Structure):
    """handles 'LVSR' resource types"""

    SUSPEND_ON_RUN = (0, 0x00001000)
    LOCKED = (0, 0x00002000)
    RUN_ON_OPEN = (0, 0x00004000)
    SAVE_FOR_PREVIOUS = (1, 0x00000004)
    SEPARATE_CODE = (1, 0x00000400)
    CLEAR_INDICATORS = (1, 0x01000000)
    AUTO_ERROR_HANDLING = (1, 0x20000000)
    BREAKPOINTS_SET = (5, 0x20000000)
    DEBUGGABLE = (5, 0x40000200)
    BREAKPOINT_COUNT_INDEX = 28

    def __init__(self):
        super().__init__(
            "version",
            Version(),
            "flags",
            Bytes(),
        )

    def __get_flag_set(self, flag: int) -> int:
        value = UInt32()

        if flag * value.size() + value.size() > self.flags.size():
            return None

        return value.from_bytes(self.flags.value, flag * value.size()).value

    def __set_flag_set(self, flag: int, new_value: int):
        value = UInt32(new_value)

        if flag * value.size() + value.size() <= self.flags.size():
            prefix = self.flags.value[: flag * value.size()]
            suffix = self.flags.value[(flag + 1) * value.size() :]
            self.flags.value = prefix + value.to_bytes() + suffix

    def __flag_value(self, flag: int, mask: int, new_value: bool) -> bool:
        flag_value = self.__get_flag_set(flag)
        old_value = None if flag_value is None else (flag_value & mask != 0)

        if new_value is not None:
            if new_value:
                flag_value |= mask
            else:
                flag_value &= ~mask

            self.__set_flag_set(flag, flag_value)

        return old_value

    def breakpoint_count(self):
        """returns the number of breakpoints or None if not supported"""
        return self.__get_flag_set(TypeLVSR.BREAKPOINT_COUNT_INDEX)

    def debuggable(self, value: bool = None) -> bool:
        """VI was marked as debuggable"""
        return self.__flag_value(*TypeLVSR.DEBUGGABLE, value)

    def clear_indicators(self, value: bool = None) -> bool:
        """When run the unwired indicators will be cleared"""
        return self.__flag_value(*TypeLVSR.CLEAR_INDICATORS, value)

    def run_on_open(self, value: bool = None) -> bool:
        """Was this VI marked as run-on-open"""
        return self.__flag_value(*TypeLVSR.RUN_ON_OPEN, value)

    def suspend_on_run(self, value: bool = None) -> bool:
        """Was this VI marked as suspend on run"""
        return self.__flag_value(*TypeLVSR.SUSPEND_ON_RUN, value)

    def locked(self, value: bool = None) -> bool:
        """Was this VI locked (possibly with password)"""
        return self.__flag_value(*TypeLVSR.LOCKED, value)

    def auto_error_handling(self, value: bool = None) -> bool:
        """Was this VI saved with auto error handling turned on"""
        return self.__flag_value(*TypeLVSR.AUTO_ERROR_HANDLING, value)

    def saved_for_previous(self, value: bool = None) -> bool:
        """Was this VI saved for previous version"""
        return self.__flag_value(*TypeLVSR.SAVE_FOR_PREVIOUS, value)

    def has_breakpoints(self, value: bool = None) -> bool:
        """Was this VI saved for previous version"""
        return self.__flag_value(*TypeLVSR.BREAKPOINTS_SET, value)

    def separate_code(self, value: bool = None) -> bool:
        """Was this VI saved with code separate"""
        return self.__flag_value(*TypeLVSR.SEPARATE_CODE, value)

    # pylint: disable=attribute-defined-outside-init
    def from_bytes(self, data: bytes, offset: int = 0):
        """Take raw bytes from the file and interpret them.
        offset - the offset in data to start parsing the bytes
        """
        self.flags = Bytes(byte_count=len(data) - offset - self.version.size())
        return super().from_bytes(data, offset)

    def __repr__(self) -> str:
        return f"TypeLVSR({self.to_string()})"


class Typevers(Structure):
    """handles 'vers' resource types"""

    ENGLISH = 0
    FRENCH = 1
    GERMAN = 3
    JAPANESE = 14
    KOREAN = 23
    CHINESE = 33
    LANGUAGES = [0, 1, 3, 14, 23, 33]

    def __init__(
        self,
        version: any = None,
        language: int = None,
        text: bytes = None,
        name: bytes = None,
    ):
        super().__init__(
            "version",
            version if isinstance(version, Version) else Version(version),
            "language",
            UInt16(language or Typevers.ENGLISH),
            "text",
            PString(text or b""),
            "name",
            PString(name or b""),
        )
        assert self.language in Typevers.LANGUAGES

    def from_bytes(self, data: bytes, offset: int = 0):
        """Take raw bytes from the file and interpret them.
        offset - the offset in data to start parsing the bytes
        """
        super().from_bytes(data, offset)
        assert self.language in Typevers.LANGUAGES
        return self

    def to_string(self):
        """Get a string representation of the vers resource information"""
        return (
            "{"
            + f"version={self.version.to_string()}, "
            + f"language={self.language}, "
            + f"text='{self.text}', "
            + f"name='{self.name}'"
            + "}"
        )

    def __repr__(self) -> str:
        return f"Typevers({self.to_string()})"

    def from_value(self, description: any):
        super().from_value(description)
        assert self.language in Typevers.LANGUAGES
        return self

    def __lt__(self, other):
        if isinstance(other, Typevers):
            return self.version < other.version

        return self.version < other

    def __eq__(self, other):
        if isinstance(other, Typevers):
            return self.version == other.version

        return self.version == other



"""
    Parses a LabVIEW resource file
"""


import argparse
import json
import sys

import pylavi.file


def parse_args(command_line=None):
    """Interpret command line arguments"""
    command_line = command_line or sys.argv[1:]
    parser = argparse.ArgumentParser(description="Converts LabVIEW Resources to text")
    parser.add_argument(
        "-d",
        "--disassemble",
        dest="disassemble",
        action="store_true",
        help="Disassemble the resource file into text",
    )
    parser.add_argument("input_file")
    parser.add_argument(
        "-o", "--out", dest="output_file", type=str, help="Path to output the results"
    )
    args = parser.parse_args(args=command_line)
    return args


def main(args=None):
    """run the (dis)assembler"""
    args = args or parse_args()
    resources = pylavi.file.Resources.load(args.input_file)
    output = {
        "path": args.input_file,
        "type": resources.file_type.to_string(),
        "creator": resources.file_creator.to_string(),
        "resources": {},
    }

    for resource_type in resources.types():
        output["resources"][resource_type] = []

        for identifier, name, data in resources.get_resources(resource_type):
            output["resources"][resource_type].append(
                {
                    "id": identifier,
                    "name": name.decode("ascii") if name else name,
                    "data": data.hex(),
                }
            )

    if args.output_file:
        print(output)
        with open(args.output_file, "w", encoding="utf-8") as output_file:
            json.dump(output, output_file, indent=2, sort_keys=True)
    else:
        json.dump(output, sys.stdout, indent=2, sort_keys=True)


if __name__ == "__main__":
    main(parse_args())



"""
    Parses a LabVIEW resource file
"""


from pylavi.data_types import Structure, Array, FourCharCode, PString, UInt16, UInt32


# pylint: disable=no-member
class Header(Structure):
    """Resource header."""

    FILE_TYPES = [
        "LVIN",  # VI
        "LVAR",  # LLB
        "LVCC",  # template VI?
        "LMNU",  # menu
        "LVRS",  # menu
        "LVSB",  # lsb
        "\\x00\\x00\\x00\\x00",  # menu
        "iUWl",  # LabWindows/CVI user interface resource file .uir
    ]
    FILE_CREATORS = [
        "LBVW",  # usual
        "WLin",  # LabWindows/CVI user interface resource file .uir
        "\\x00\\x00\\x00\\x00",  # older .uir and .mnu files
    ]
    VERSION = 3
    SIGNATURE = FourCharCode().from_bytes(b"RSRC")
    CORRUPTION_CHECK = b"\r\n"

    def __init__(self, file_type: str = None, file_creator: str = None):
        file_type = file_type or Header.FILE_TYPES[0]
        file_creator = file_creator or Header.FILE_CREATORS[0]
        super().__init__(
            "file_format",
            FourCharCode(),
            "corruption_check",
            UInt16(),
            "format_version",
            UInt16(),
            "file_type",
            FourCharCode(file_type),
            "file_creator",
            FourCharCode(file_creator),
            "metadata_offset",
            UInt32(),
            "metadata_size",
            UInt32(),
            "data_offset",
            UInt32(),
            "data_size",
            UInt32(),
        )

    def to_string(self):
        """String representation of the header"""
        return (
            "{"
            + f"file_type={self.file_type.to_string()}, "
            + f"file_creator={self.file_creator.to_string()}, "
            + f"metadata_offset={self.metadata_offset}, metadata_size={self.metadata_size}, "
            + f"data_offset={self.data_offset}, data_size={self.data_size}"
            + "}"
        )

    def __repr__(self) -> str:
        return f"Header({self.to_string()})"

    def validate(self, file_size: int = None):
        """ensured the header data makes sense"""
        assert (
            self.file_format == Header.SIGNATURE
        ), f"Invalid Signature {self.file_format}"
        assert self.corruption_check == b"\r\n", f"Corrupt {self.corruption_check}"
        assert self.format_version == Header.VERSION, f"Version {self.format_version}"
        assert (
            self.file_type.to_value() in Header.FILE_TYPES
        ), f"Type {[self.file_type]}"
        assert (
            self.file_creator.to_value() in Header.FILE_CREATORS
        ), f"Creator {[self.file_creator]}"
        assert (
            self.size() == self.data_offset
        ), f"post-header gap {self.data_offset - self.size()}"
        assert (
            self.data_offset + self.data_size == self.metadata_offset
        ), f"post-data gap {self.metadata_offset - (self.data_offset + self.data_size)}"
        assert self.data_size.value % 4 == 0
        assert (
            file_size is None or file_size >= self.metadata_offset + self.metadata_size
        ), (
            f"file_size={file_size} expected="
            + f"{self.metadata_offset + self.metadata_size}\n"
            + f"\t metadata offset = {self.metadata_offset}\n"
            + f"\t metadata size = {self.metadata_size}"
        )
        minimum_metadata_size = (
            Header().size()
            + MetadataHeader().size()
            + UInt32().size()
            + TypeInfo().size()
        )
        assert (
            self.metadata_size.value >= minimum_metadata_size
        ), f"metadata size too small {self.metadata_size}"
        assert self.metadata_offset == (self.data_offset + self.data_size), (
            f"metadata should be right after data: metadata = {self.metadata_offset}"
            + f" data end = {self.data_offset + self.data_size}"
        )
        return self


class MetadataHeader(Structure):
    """header for the metadata section"""

    def __init__(self):
        header_size = Header().size()
        super().__init__(
            "unused_8",
            UInt32(0),
            "unused_16",
            UInt32(0),
            "file_header_size",
            UInt32(header_size),
            "metadata_header_size",
            UInt32(),
            "names_offset",
            UInt32(),
        )
        self.metadata_header_size.value = header_size + self.size()

    def to_string(self):
        """String representation of the header"""
        return (
            "{"
            + f"unused_8 = {self.unused_8}, "
            + f"unused_16 = {self.unused_16}, "
            + f"file_header_size = {self.file_header_size}, "
            + f"metadata_header_size = {self.metadata_header_size}, "
            + f"names_offset = {self.names_offset}"
            + "}"
        )

    def __repr__(self) -> str:
        return f"MetadataHeader({self.to_string()})"

    def validate(self, file_header: Header, file_size: int):
        """validate that the data makes sense"""
        assert (
            self.file_header_size == Header().size()
        ), f"types header incorrect size: {self.file_header_size}"
        assert (
            self.metadata_header_size == Header().size() + self.size()
        ), f"type info not where it should be {self.metadata_header_size}"
        assert file_size >= (
            file_header.metadata_offset.value + self.names_offset.value
        ), (
            f"File too small {file_size} <"
            + f" {file_header.metadata_offset + self.names_offset}"
            + f" file_header.metadata_offset = {file_header.metadata_offset}"
            + f" names_offset = {self.names_offset}"
        )
        return self


class TypeInfo(Structure):
    """Structure for the list of resource types."""

    def __init__(
        self,
        resource_type: str = "\0\0\0\0",
        resource_count: int = 1,
        list_offset: int = 0,
    ):
        super().__init__(
            "resource_type",
            FourCharCode(resource_type),
            "resource_count",
            UInt32(resource_count - 1),
            "list_offset",
            UInt32(list_offset),
        )

    def to_string(self):
        """get string values"""
        return (
            "TypeInfo{"
            + f"resource_type = {self.resource_type.to_string()}, "
            + f"resource_count = {self.resource_count} (+1), "
            + f"list_offset = {self.list_offset}"
            + "}"
        )

    def __repr__(self) -> str:
        return f"TypeInfo({self.to_string()})"

    def number_of_resources(self) -> int:
        """Get the number of resources for the given type"""
        return self.resource_count + 1


class TypeList(Array):
    """Fixed size list of types"""

    def __init__(self, *elements, length: int = 0):
        assert not elements or all(isinstance(e, TypeInfo) for e in elements)
        super().__init__(TypeInfo, *elements, data_count=length)

    def from_bytes(self, data: bytes, offset: int = 0):
        """fill in data from bytes"""
        length = UInt32().from_bytes(data, offset) + 1
        offset += UInt32().size()
        assert offset + length * TypeInfo().size() <= len(data), [
            offset,
            length,
            offset + length * TypeInfo().size(),
            len(data),
            data,
        ]
        self.set_length(length)
        super().from_bytes(data, offset)
        return self

    def to_bytes(self) -> bytes:
        """get the binary version"""
        return UInt32(self.length() - 1).to_bytes() + super().to_bytes()

    def __repr__(self):
        return f"TypeList({', '.join(repr(v) for v in self.value)})"


class ResourceMetadata(Structure):
    """Structure for the RsrcEntry Entry in the resource map."""

    NO_NAME = 0xFFFFFFFF

    def __init__(
        self, res_id: int = None, name_offset: int = None, data_offset: int = None
    ):
        super().__init__(
            "resource_id",
            UInt32(res_id),
            "name_offset",
            UInt32(ResourceMetadata.NO_NAME if name_offset is None else name_offset),
            "unused_8",
            UInt32(0),
            "data_offset",
            UInt32(data_offset),
            "unused_16",
            UInt32(0),
        )
        assert (
            self.name_offset.value is None
            or self.name_offset.value % 4 == 0
            or self.name_offset.value == ResourceMetadata.NO_NAME
        )
        assert self.data_offset.value is None or self.data_offset.value % 4 == 0

    def to_string(self):
        """Convert to a string"""
        return (
            "{"
            + f"resource_id = {self.resource_id}, "
            + f"unused_8 = {self.unused_8}, "
            + f"unused_16 = {self.unused_16}, "
            + f"name_offset = {self.name_offset}, "
            + f"data_offset = {self.data_offset}"
            + "}"
        )

    def __repr__(self) -> str:
        return f"ResourceMetadata({self.to_string()})"


class ResourceList(Array):
    """Fixed size list of types"""

    def __init__(self, *elements, length: int = 0):
        assert not elements or all(isinstance(e, ResourceMetadata) for e in elements)
        super().__init__(ResourceMetadata, *elements, data_count=length)

    def __repr__(self):
        return f"ResourceList({', '.join(repr(v) for v in self.value)})"


class Resources:
    """Resources from a LabVIEW resource file"""

    EXTENSIONS = [
        ".vi",
        ".vit",
        ".ctl",
        ".ctt",
        ".llb",
        ".vim",
        ".mnu",
        ".uir",
        ".lsb",
        ".rtexe",
        ".gbl",
        ".glb",
    ]

    def __init__(
        self, file_type: str = None, file_creator: str = None, description: list = None
    ):
        self.file_type = file_type
        self.file_creator = file_creator
        self.__resources = description

    def types(self) -> [str]:
        """Get the types contained in the resource file"""
        return list(t[0] for t in self.__resources)

    def count_type(self, resource_type: str) -> int:
        """Get the number of resources of the given type"""
        return sum(len(t[1]) for t in self.__resources if t[0] == resource_type)

    def get_ids(self, resource_type: str) -> [int]:
        """Get the id of the resources"""
        return list(
            r[0].value for t in self.__resources if t[0] == resource_type for r in t[1]
        )

    def get_names(self, resource_type: str) -> [str]:
        """Get the names of the resources"""
        return list(
            r[1] for t in self.__resources if t[0] == resource_type for r in t[1]
        )

    def get_resources(self, resource_type: str) -> [(int, str, bytes)]:
        """Gets all resources of a given type with their name and id"""
        return list(
            (r[0].value, r[1], r[2])
            for t in self.__resources
            if t[0] == resource_type
            for r in t[1]
        )

    def get_resource(
        self, resource_type: str, resource_id: int = None, name: str = None
    ) -> bytes:
        """Get a resource either by name of id"""
        assert resource_id is not None or name is not None

        if name is None:
            return list(
                r[2]
                for t in self.__resources
                if t[0] == resource_type
                for r in t[1]
                if r[0] == resource_id
            )[0]

        with_name = list(
            r[2]
            for t in self.__resources
            if t[0] == resource_type
            for r in t[1]
            if r[1] == name.encode("ascii")
        )
        return None if not with_name else with_name[0]

    @staticmethod
    def __load_file_header(contents: bytes) -> Header:
        return Header().from_bytes(contents).validate(len(contents))

    @staticmethod
    def __validate_2nd_file_header(contents: bytes, header: Header) -> int:
        offset = header.metadata_offset.value
        second_header = Header().from_bytes(contents, offset).validate(len(contents))
        assert (
            second_header == header
        ), f"headers don't match {header} vs {second_header}"
        return offset + second_header.size()

    @staticmethod
    def __load_metadata_header(
        contents: bytes, header: Header, offset: int
    ) -> (int, MetadataHeader):
        metadata_header = (
            MetadataHeader()
            .from_bytes(contents, offset)
            .validate(header, len(contents))
        )
        offset += metadata_header.size()
        return offset, metadata_header

    @staticmethod
    def __load_typelist(contents: bytes, offset: int) -> any:
        return TypeList().from_bytes(contents, offset)

    @staticmethod
    def __load_resource_name(
        header: Header,
        metadata_header: MetadataHeader,
        resource_info: ResourceMetadata,
        contents: bytes,
    ) -> str:
        if resource_info.name_offset != ResourceMetadata.NO_NAME:
            name_offset = (
                header.metadata_offset.value
                + metadata_header.names_offset.value
                + resource_info.name_offset.value
            )
            name = PString().from_bytes(contents, name_offset)
            assert len(name.value) > 0, f"name_size = {len(name.value)}"
            return name.value

        return None

    @staticmethod
    def __load_resource_data(header, resource_info, contents):
        data_offset = header.data_offset + resource_info.data_offset
        data_size = UInt32().from_bytes(contents, data_offset)
        data_offset += data_size.size()
        offset_past_data = data_offset + data_size.value
        return contents[data_offset:offset_past_data]

    @staticmethod
    def load(path: str):
        """Loads the resources from the LabVIEW resource file"""
        with open(path, "rb") as resource_file:
            contents = resource_file.read()

        assert len(contents) >= 2 * Header().size(), "File too small"
        header = Resources.__load_file_header(contents)
        offset = Resources.__validate_2nd_file_header(contents, header)
        offset, metadata_header = Resources.__load_metadata_header(
            contents, header, offset
        )
        data_types = Resources.__load_typelist(contents, offset)
        resource_types = []
        last_offset = 0

        for entry in data_types:
            assert entry.resource_count < int(
                header.metadata_size.value / ResourceMetadata().size()
            ), "Corrupted"
            assert (
                entry.list_offset > last_offset
            ), f"Unordered type table {entry.list_offset} vs {last_offset}"
            last_offset = entry.list_offset
            resource_list = ResourceList(length=entry.number_of_resources())
            offset = header.metadata_offset
            offset += Header().size() + MetadataHeader().size()
            offset += entry.list_offset.value
            resource_list.from_bytes(contents, offset)
            resources = [
                (
                    r.resource_id,
                    Resources.__load_resource_name(
                        header, metadata_header, r, contents
                    ),
                    Resources.__load_resource_data(header, r, contents),
                )
                for r in resource_list
            ]
            resource_types.append((entry.resource_type.to_string(), resources))

        return Resources(
            file_type=header.file_type,
            file_creator=header.file_creator,
            description=resource_types,
        )



"""
    Checks LabVIEW resource file(s) for given criteria
"""


import argparse
import fnmatch
import os
import queue
import sys
import threading
import time
import types

import yaml

from pylavi.file import Resources
from pylavi.resource_types import Typevers, TypeLVSR, TypeBDPW
from pylavi.data_types import Version, Path


MAX_FILES_TO_QUEUE = 1000


class Problem:
    """Describe a problem found with a file"""

    def __init__(self, path, message, other=None):
        self.path = path
        self.message = message
        self.other = other

    def to_string(self):
        """Displayable message about the problem"""
        return f"{self.message}: {self.path}" + (
            ("\n\t" + "\n\t".join(self.other)) if self.other else ""
        )

    def __str__(self):
        return self.to_string()


def add_version_options(parser):
    """adds options related to LabVIEW version that saved the file"""
    parser.add_argument(
        "-l", "--lt", type=str, help="LabVIEW version must be less than this"
    )
    parser.add_argument(
        "-g", "--gt", type=str, help="LabVIEW version must be greater than this"
    )
    parser.add_argument("-e", "--eq", type=str, help="LabVIEW version must this")
    parser.add_argument(
        "-r",
        "--no-release",
        action="store_true",
        help="LabVIEW version must not be release",
    )
    parser.add_argument(
        "-b", "--no-beta", action="store_true", help="LabVIEW version must not be beta"
    )
    parser.add_argument(
        "-a",
        "--no-alpha",
        action="store_true",
        help="LabVIEW version must not be alpha",
    )
    parser.add_argument(
        "-d",
        "--no-development",
        action="store_true",
        help="LabVIEW version must not be development",
    )
    parser.add_argument(
        "-i",
        "--no-invalid",
        action="store_true",
        help="LabVIEW version must be a valid phase",
    )


def add_flag_options(parser):
    """add options for checking flags"""
    parser.add_argument("-c", "--no-code", action="count", help="Saved without code")
    parser.add_argument("--code", action="count", help="Saved with code")
    parser.add_argument(
        "--breakpoints", action="store_true", help="Not saved with breakpoints"
    )
    parser.add_argument(
        "--locked", action="count", help="File is locked (maybe with password)"
    )
    parser.add_argument("--not-locked", action="count", help="File is not locked")
    parser.add_argument(
        "--password-match", type=str, help="Ensure the password is exactly"
    )
    parser.add_argument(
        "--password", action="count", help="File is locked with a password"
    )
    parser.add_argument(
        "--no-password", action="count", help="File is not saved with a password"
    )
    parser.add_argument(
        "--clear-indicators", action="count", help="VI will clear indicators on run"
    )
    parser.add_argument(
        "--no-clear-indicators",
        action="count",
        help="VI will not clear indicators on run",
    )
    parser.add_argument("--run-on-open", action="count", help="VI will run when opened")
    parser.add_argument(
        "--no-run-on-open", action="count", help="VI will not run when opened"
    )
    parser.add_argument(
        "--suspend-on-run", action="count", help="VI will suspend on run"
    )
    parser.add_argument(
        "--no-suspend-on-run", action="count", help="VI will not suspend on run"
    )
    parser.add_argument("--debuggable", action="count", help="VI is debuggable")
    parser.add_argument("--not-debuggable", action="count", help="VI is not debuggable")
    parser.add_argument(
        "--no-absolute-path",
        action="store_true",
        help="Does not reference links by absolute path",
    )
    parser.add_argument(
        "--autoerror",
        action="store_true",
        help="Not saved with auto error handling turned on",
    )


def patch_up_args(args):
    """from the raw args patch them up to be useful"""
    args.path = args.path or ([] if args.config else ["."])
    args.skip = args.skip or []
    args.quiet = args.quiet or 0
    args.verbose = args.verbose or 0
    args.verbose -= args.quiet
    args.quiet = -1 * args.verbose
    args.no_code = args.no_code or 0
    args.code = args.code or 0
    args.no_password = args.no_password or 0
    args.password = args.password or 0
    args.not_locked = args.not_locked or 0
    args.locked = args.locked or 0
    args.not_debuggable = args.not_debuggable or 0
    args.debuggable = args.debuggable or 0
    args.no_clear_indicators = args.no_clear_indicators or 0
    args.clear_indicators = args.clear_indicators or 0
    args.no_run_on_open = args.no_run_on_open or 0
    args.run_on_open = args.run_on_open or 0
    args.no_suspend_on_run = args.no_suspend_on_run or 0
    args.suspend_on_run = args.suspend_on_run or 0
    has_comparison = args.lt or args.gt or args.eq
    has_phase = (
        args.no_release
        or args.no_beta
        or args.no_alpha
        or args.no_development
        or args.no_invalid
    )
    has_code = args.no_code > 0 or args.code > 0
    has_locked = args.locked > 0 or args.not_locked > 0
    has_password = args.password > 0 or args.no_password > 0
    has_debuggable = args.debuggable > 0 or args.not_debuggable > 0
    has_binary = has_code or has_locked or has_password or has_debuggable
    has_other = (
        args.autoerror or args.breakpoints or args.password_match or args.path_length
    )

    if (
        not has_comparison
        and not has_phase
        and not has_binary
        and not has_other
        and not args.config
    ):
        args.no_beta = True
        args.no_alpha = True
        args.no_development = True
        args.no_invalid = True
        args.path_length = 260
        args.no_absolute_path = True

    if not args.extension:
        args.extension = Resources.EXTENSIONS

    return args


def parse_args(command_line=None):
    """Interpret command line arguments"""
    command_line = command_line or sys.argv[1:]
    parser = argparse.ArgumentParser(description="Validates LabVIEW resource files")
    add_version_options(parser)
    add_flag_options(parser)

    parser.add_argument(
        "--path-length",
        type=int,
        help="Maximum number of characters for the path",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        action="append",
        help="Path to scan for files (or a file path) (defaults to current directory)",
    )
    parser.add_argument(
        "-s",
        "--skip",
        type=str,
        action="append",
        help="Path to not scan for files (or a file to ignore)",
    )
    parser.add_argument(
        "-x",
        "--extension",
        type=str,
        action="append",
        help="File extensions to evaluate (defaults to all known)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        help="Reduce the output (multiple times reduces output more)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Increase the output (multiple times increases output more)",
    )
    parser.add_argument(
        "--config",
        type=str,
        help="Path to a .yaml file with collections of arguments",
    )
    return patch_up_args(parser.parse_args(args=command_line))


def find_files(found_queue, argsset):
    """Find files in given paths and put them in the queue"""
    for args in argsset:
        for path in args.path:
            if os.path.isfile(path):
                found_queue.put((args, path))
            else:
                for root, _, files in os.walk(path):
                    for file in files:
                        found_queue.put((args, os.path.join(root, file)))

    found_queue.put(None)


def start_finding_files(args):
    """Start the find_files function running in the background finding files"""
    found = queue.Queue(MAX_FILES_TO_QUEUE)
    thread = threading.Thread(target=find_files, args=(found, args), daemon=True)
    thread.start()
    return found


def find_paths(data):
    """Search a block of data for potential Path data in it"""
    paths = []

    while b"PTH" in data:
        try:
            data = data[data.find(b"PTH") :]
            paths.append(Path().from_bytes(data))
            data = data[paths[-1].size() :]
        except AssertionError:  # Just a random PTH in the resource, not a path
            data = data[1:]

    return paths


# pylint: disable=too-many-return-statements
def validate_run(args, save_record: TypeLVSR, problems: list, next_path: str) -> bool:
    """validate run flags"""

    if (
        args.clear_indicators - args.no_clear_indicators < 0
        and save_record.clear_indicators()
    ):
        problems.append(Problem(next_path, "Clear Indicators is on"))
        return True

    if (
        args.clear_indicators - args.no_clear_indicators > 0
        and not save_record.clear_indicators()
    ):
        problems.append(Problem(next_path, "Clear Indicators is off"))
        return True

    if args.run_on_open - args.no_run_on_open > 0 and not save_record.run_on_open():
        problems.append(Problem(next_path, "Run on open is off"))
        return True

    if args.run_on_open - args.no_run_on_open < 0 and save_record.run_on_open():
        problems.append(Problem(next_path, "Run on open is on"))
        return True

    if (
        args.suspend_on_run - args.no_suspend_on_run > 0
        and not save_record.suspend_on_run()
    ):
        problems.append(Problem(next_path, "Suspend on run is off"))
        return True

    if (
        args.suspend_on_run - args.no_suspend_on_run < 0
        and save_record.suspend_on_run()
    ):
        problems.append(Problem(next_path, "Suspend on run is on"))
        return True

    return False


def validate_code(args, save_record: TypeLVSR, problems: list, next_path: str) -> bool:
    """Validates if the separate code flags matches the command line arguments"""

    if args.code - args.no_code > 0 and save_record.separate_code():
        problems.append(Problem(next_path, "Code is separate from VI"))
        return True

    if args.code - args.no_code < 0 and not save_record.separate_code():
        problems.append(Problem(next_path, "Code is not separate from VI"))
        return True

    if args.debuggable - args.not_debuggable > 0 and not save_record.debuggable():
        problems.append(Problem(next_path, "Not debuggable"))
        return True

    if args.debuggable - args.not_debuggable < 0 and save_record.debuggable():
        problems.append(Problem(next_path, "Debuggable"))
        return True

    return False


def validate_locked_password(
    args, save_record: TypeLVSR, password: TypeBDPW, problems: list, next_path: str
) -> bool:
    """Validates if the separate code flags matches the command line arguments"""
    if args.password_match and not password.password_matches(args.password_match):
        problems.append(Problem(next_path, f"Password is not '{args.password_match}'"))
        return True

    if args.password - args.no_password > 0:
        if not save_record.locked() or not password.has_password():
            problems.append(Problem(next_path, "No password"))
            return True

    if args.password - args.no_password < 0:
        if save_record.locked() and password.has_password():
            problems.append(Problem(next_path, "Has password"))
            return True

    if args.locked - args.not_locked > 0:
        if not save_record.locked():
            problems.append(Problem(next_path, "Not locked"))
            return True

    if args.locked - args.not_locked < 0:
        if save_record.locked():
            problems.append(Problem(next_path, "Locked"))
            return True

    return False


def validate_version(args, versions, problems, next_path):
    """Validates the version checking logic"""
    for version in versions:
        version_string = version.to_string()
        phase = version.phase()

        if args.gt and version < args.gt:
            problems.append(
                Problem(next_path, f"VI saved in LabVIEW {version_string} < {args.gt}")
            )
            break

        if args.lt and not version < args.lt:
            problems.append(
                Problem(next_path, f"VI saved in LabVIEW {version_string} > {args.lt}")
            )
            break

        if args.eq and not version == args.eq:
            problems.append(
                Problem(
                    next_path, f"VI saved in LabVIEW {version_string} is not {args.eq}"
                )
            )
            break

        if args.no_release and phase == Version.RELEASE:
            problems.append(
                Problem(
                    next_path,
                    f"VI saved in LabVIEW {version_string} is a release version",
                )
            )
            break

        if args.no_beta and phase == Version.BETA:
            problems.append(
                Problem(
                    next_path, f"VI saved in LabVIEW {version_string} is a beta version"
                )
            )
            break

        if args.no_alpha and phase == Version.ALPHA:
            problems.append(
                Problem(
                    next_path,
                    f"VI saved in LabVIEW {version_string} is an alpha version",
                )
            )
            break

        if args.no_development and phase == Version.DEVELOPMENT:
            problems.append(
                Problem(
                    next_path,
                    f"VI saved in LabVIEW {version_string} is a development version",
                )
            )
            break

        if args.no_invalid and (phase > Version.RELEASE or phase == 0):
            problems.append(
                Problem(
                    next_path,
                    f"VI saved in LabVIEW {version_string} is an invalid phase",
                )
            )
            break


def validate_links(resources: Resources, problems: list, next_path: str):
    """validates that references are not linked via absolute paths"""
    bad_paths = []

    for _, _, data in resources.get_resources("LIvi"):
        for path in find_paths(data):
            if path.get_type() in [Path.RELATIVE, Path.NOTAPATH]:
                continue

            if path.to_value() is None:
                continue

            absolute = path.is_type(Path.ABSOLUTE)
            has_elements = path.elements.length() > 0
            begin_named = not has_elements or path.elements[0].value.startswith(b"<")
            end_named = not has_elements or path.elements[0].value.endswith(b">")

            if absolute and has_elements and begin_named and end_named:
                continue

            bad_paths.append(str(path))

    if bad_paths:
        plural = "" if len(bad_paths) == 1 else "s"
        problems.append(
            Problem(next_path, f"Absolute linker path{plural} found", bad_paths)
        )
        return True

    return False


# pylint: disable=no-member
def validate(args, resources: Resources, problems: list, next_path: str):
    """Validate that the given resources file is valid"""
    version_resources = resources.get_resources("vers")
    save_record_resources = resources.get_resources("LVSR")
    password_resources = resources.get_resources("BDPW")
    assert len(save_record_resources) < 2
    versions = [Typevers().from_bytes(r[2]).version for r in version_resources]
    save_record = (
        TypeLVSR().from_bytes(save_record_resources[0][2])
        if save_record_resources
        else None
    )
    password_record = (
        TypeBDPW().from_bytes(password_resources[0][2]) if password_resources else None
    )
    problem_count = len(problems)
    invalid = False

    if save_record:
        versions.append(save_record.version)

    if not invalid and args.no_absolute_path:
        invalid = validate_links(resources, problems, next_path)

    if not invalid and args.path_length and len(next_path) > args.path_length:
        problems.append(
            Problem(next_path, f"Path length {len(next_path)} > {args.path_length}")
        )
        invalid = True

    if not invalid and args.breakpoints and save_record:
        if save_record.has_breakpoints():
            number = save_record.breakpoint_count()
            plural = "" if number == 1 else "s"
            number = "" if number is None else f"{number} "
            problems.append(Problem(next_path, f"{number} breakpoint{plural} found"))
            invalid = True

    if (
        not invalid
        and args.autoerror
        and save_record
        and save_record.auto_error_handling()
    ):
        problems.append(Problem(next_path, "Auto error handling turned on"))
        invalid = True

    if not invalid and save_record_resources:
        invalid = validate_code(args, save_record, problems, next_path)

    if not invalid and save_record_resources:
        invalid = validate_run(args, save_record, problems, next_path)

    if not invalid and save_record_resources and password_record:
        invalid = validate_locked_password(
            args, save_record, password_record, problems, next_path
        )

    if not invalid:
        validate_version(args, versions, problems, next_path)

    if len(problems) > problem_count and args.quiet < 1:
        print(f"FAIL: {problems[-1]}")


def should_skip(args, path, verbose: int):
    """Determine if the file should be skipped"""
    if os.path.splitext(path)[1] not in args.extension:
        return True

    if verbose > 2:
        print(f"path = {path}")

        for skip in [] if not args.skip else args.skip:
            print(f"\t skip = {skip} fnmatch = {fnmatch.fnmatch(path, skip)}")

    return any(fnmatch.fnmatch(path, p) for p in ([] if not args.skip else args.skip))


def find_problems(files, top_args):
    """Searchs for problems files and returns them"""
    problems = []
    file_count = 0
    start = time.time()

    while True:
        next_path = files.get()

        if next_path is None:
            break

        args, next_path = next_path
        suffix = ""

        if "name" in dir(args):
            suffix = f" from {args.name}"

        if should_skip(args, next_path, verbose=args.verbose):
            if args.verbose > 1:
                print(f"Skipping: {next_path}{suffix}")
            continue

        try:
            if args.verbose > 0:
                print(f"Validating: {next_path}{suffix}")

            file_count += 1
            validate(args, Resources.load(next_path), problems, next_path)

        except AssertionError as error:
            problems.append(Problem(next_path, error))
            if args.quiet < 2:
                print(f"FAIL: {problems[-1]}")

    if top_args.verbose > 1:
        print(
            f"{file_count} files in {time.time() - start:0.1f} seconds "
            + f"{file_count / (time.time() - start):0.0f} files/second"
        )

    return problems


def parse_config_file(args):
    """parse the config yaml file and generate args"""
    if not args.config:
        return (args,)

    with open(args.config, "r", encoding="utf-8") as config_file:
        more_args = yaml.load(config_file, Loader=yaml.Loader)

    config_dir = os.path.dirname(args.config)
    full_arg_set = {a: None for a in dir(args) if not a.startswith("_")}
    config_args = []

    for arg_set_name in more_args:
        assert set(more_args[arg_set_name]) <= set(full_arg_set), (
            f"Unknown {set(more_args[arg_set_name]) - set(full_arg_set)} "
            + f"in '{arg_set_name}' section in {args.config}"
        )
        arg_set = dict(full_arg_set)
        arg_set.update(more_args[arg_set_name])
        arg_set["name"] = arg_set_name
        arg_set["relative_to"] = config_dir

        for path_type in ["path", "skip"]:
            if arg_set.get(path_type, False):
                arg_set[path_type] = [
                    os.path.join(config_dir, p) for p in arg_set[path_type]
                ]

        config_args.append(patch_up_args(types.SimpleNamespace(**arg_set)))
        config_args[-1].verbose = args.verbose
        config_args[-1].quiet = args.quiet

    return (args, *config_args)


def main(args=None):
    """run the (dis)assembler"""
    args = parse_config_file(args or parse_args())

    if args[0].verbose > 1:
        for argset in args:
            print(
                f"{argset.name if 'name' in dir(argset) else 'CLI'}: "
                + ", ".join(
                    f"{a} = {argset.__dict__[a]}"
                    for a in dir(argset)
                    if not a.startswith("_")
                )
            )

    files = start_finding_files(args)
    problems = find_problems(files, args[0])

    if len(problems) > 0 and args[0].quiet < 3:
        print(f"{len(problems)} problems encounted")
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main(parse_args()))



""" pylavi package info
    See: https://setuptools.pypa.io/en/latest/setuptools.html#command-reference
    See: https://betterscientificsoftware.github.io/python-for-hpc/tutorials/python-pypi-packaging/
"""


import os
import re
from setuptools import setup, find_packages


VERSION_PATTERN = re.compile(
    r"^\s*__version__\s*=\s*['\"](\d+\.\d+\.\d+)['\"]", re.MULTILINE
)


def get_file_contents(*filename):
    """Get the contents of a file relative to the directory setup.py is in"""
    this_dir = os.path.dirname(os.path.abspath(__file__))
    readme_path = os.path.join(this_dir, *filename)

    if os.path.isfile(readme_path):
        with open(readme_path, "r", encoding="utf-8") as readme:
            return readme.read().strip()

    return ""


def get_version():
    """single-source the version by reading the __init__.py file"""
    metadata = get_file_contents("pylavi", "__init__.py")
    has_version = VERSION_PATTERN.search(metadata)

    if not has_version:
        raise RuntimeError("Unable to find version string.")

    return has_version.group(1)


setup(
    name="pylavi",
    version=get_version(),
    description="Python LabVIEW File Parser",
    long_description=get_file_contents("README.md"),
    long_description_content_type="text/markdown",
    author="marcpage",
    author_email="MarcAllenPage@gmail.com",
    install_requires=['PyYAML'],
    python_requires="",
    url="https://github.com/marcpage/pylavi",
    license="Unlicense",
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: The Unlicense (Unlicense)",
        "Programming Language :: Python :: 3.9",
        "Topic :: Utilities",
        "Topic :: Software Development",
    ],
    keywords=(
        "labview vi mnu ctl instruments parser extractor "
        + "reverse-engineering development"
    ),
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "lv_assembler = pylavi.assembler:main",
            "vi_validate = pylavi.validate:main",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/marcpage/pylavi/issues",
        "Source": "https://github.com/marcpage/pylavi/",
    },
)
